/*
 *  The contents of this file are subject to the Initial
 *  Developer's Public License Version 1.0 (the "License");
 *  you may not use this file except in compliance with the
 *  License. You may obtain a copy of the License at
 *  http://www.ibphoenix.com/main.nfs?a=ibphoenix&page=ibp_idpl.
 *
 *  Software distributed under the License is distributed AS IS,
 *  WITHOUT WARRANTY OF ANY KIND, either express or implied.
 *  See the License for the specific language governing rights
 *  and limitations under the License.
 *
 *  The Original Code was created by Dmitry Yemanov
 *  for the Firebird Open Source RDBMS project.
 *
 *  Copyright (c) 2014 Dmitry Yemanov <dimitr@firebirdsql.org>
 *  and all contributors signed below.
 *
 *  All Rights Reserved.
 *  Contributor(s): ______________________________________.
 */


#ifndef JRD_REPLICATION_LOG_H
#define JRD_REPLICATION_LOG_H

#include "../common/classes/array.h"
#include "../common/classes/semaphore.h"
#include "../common/os/guid.h"
#include "../common/isc_s_proto.h"
#include "../jrd/repl/ReplicationUtils.h"

struct LogHeader
{
	char hdr_signature[10];
	USHORT hdr_version;
	ULONG hdr_state;
	Firebird::Guid hdr_guid;
	FB_UINT64 hdr_sequence;
	ULONG hdr_protocol;
	ULONG hdr_length;
};

const char LOG_SIGNATURE[] = "FBREPLLOG";

const USHORT LOG_VERSION_1 = 1;
const USHORT LOG_CURRENT_VERSION = LOG_VERSION_1;

const ULONG LOG_STATE_FREE = 0;
const ULONG LOG_STATE_USED = 1;
const ULONG LOG_STATE_FULL = 2;
const ULONG LOG_STATE_ARCH = 3;

class ReplicationLog : protected Firebird::PermanentStorage, private ErrorLogger, public Firebird::IpcObject
{
protected:
	struct State : public Firebird::MemoryHeader
	{
		ULONG version;
		time_t timestamp;
		ULONG segmentCount;
		FB_UINT64 sequence;
		FB_UINT64 lockAcquires;
		FB_UINT64 lockBlocks;
		ULONG pidLower;
		ULONG pidUpper;
		int pids[1];
	};

	class LockGuard
	{
	public:
		LockGuard(ReplicationLog* log)
			: m_log(log)
		{
			m_log->lockState();
		}

		~LockGuard()
		{
			if (m_log)
				m_log->unlockState();
		}

		void release()
		{
			if (m_log)
			{
				m_log->unlockState();
				m_log = NULL;
			}
		}

	private:
		ReplicationLog* m_log;
	};

	class LockCheckout
	{
	public:
		LockCheckout(ReplicationLog* log)
			: m_log(log)
		{
			m_log->unlockState();
		}

		~LockCheckout()
		{
			m_log->lockState();
		}

	private:
		ReplicationLog* m_log;
	};

	class Segment : public Firebird::RefCounted
	{
	public:
		~Segment();

		static Segment* create(MemoryPool& pool,
							   const Firebird::PathName& prefix,
							   const Firebird::PathName& dir);

		static Segment* open(MemoryPool& pool,
							 const Firebird::PathName& filename);

		void init(FB_UINT64 sequence, const Firebird::Guid& guid);
		bool validate(const Firebird::Guid& guid) const;
		void append(ULONG length, const UCHAR* data, const Firebird::TimeStamp& ts);
		void copyTo(const Firebird::PathName& filename) const;

		ULONG getLength() const
		{
			return m_header->hdr_length;
		}

		FB_UINT64 getSequence() const
		{
			return m_header->hdr_sequence;
		}

		ULONG getState() const
		{
			return m_header->hdr_state;
		}

		void setState(ULONG state);

		void truncate();

		Firebird::PathName getFileName() const;
		Firebird::PathName getArchiveName() const;

	private:
		Segment(MemoryPool& pool, const Firebird::PathName& filename, int handle);

		void flushHeader();

		Firebird::PathName m_filename;
		int m_handle;
		LogHeader* m_header;

#ifdef WIN_NT
		HANDLE m_mapping;
#endif
	};

public:
	ReplicationLog(MemoryPool& pool, ReplicationConfig* config,
				   const Firebird::PathName& database,
				   const Firebird::Guid& guid, bool async);
	~ReplicationLog();

	void write(ULONG length, const UCHAR* data);

	void bgArchiver();

protected:
	void lockState();
	void unlockState();

	void linkSelf();
	bool unlinkSelf();

	bool initialize(Firebird::SharedMemoryBase* shmem, bool init);
	void mutexBug(int osErrorCode, const char* text);

	void initSegment(Segment* segment)
	{
		segment->init(++m_state->getHeader()->sequence, m_guid);
	}

	bool validateSegment(const Segment* segment)
	{
		return segment->validate(m_guid);
	}

	void initSegments();
	void clearSegments();

	Segment* getSegment(ULONG segment);
	Segment* createSegment();

	bool archiveExecute(const Segment*);
	bool archiveSegment(Segment*);

	void switchActiveSegment();

	Firebird::RefPtr<ReplicationConfig> m_config;
	Firebird::Array<Segment*> m_segments;
	Firebird::AutoPtr<Firebird::SharedMemory<State> > m_state;
	Firebird::Guid m_guid;

	Firebird::Semaphore m_startupSemaphore;
	Firebird::Semaphore m_cleanupSemaphore;
	Firebird::Semaphore m_workingSemaphore;

	bool m_shutdown;
	bool m_async;
};

#endif // JRD_REPLICATION_LOG_H

