/*
 *  The contents of this file are subject to the Initial
 *  Developer's Public License Version 1.0 (the "License");
 *  you may not use this file except in compliance with the
 *  License. You may obtain a copy of the License at
 *  http://www.ibphoenix.com/main.nfs?a=ibphoenix&page=ibp_idpl.
 *
 *  Software distributed under the License is distributed AS IS,
 *  WITHOUT WARRANTY OF ANY KIND, either express or implied.
 *  See the License for the specific language governing rights
 *  and limitations under the License.
 *
 *  The Original Code was created by Dmitry Yemanov
 *  for the Firebird Open Source RDBMS project.
 *
 *  Copyright (c) 2014 Dmitry Yemanov <dimitr@firebirdsql.org>
 *  and all contributors signed below.
 *
 *  All Rights Reserved.
 *  Contributor(s): ______________________________________.
 */

#include "firebird.h"
#include "../common/classes/GenericMap.h"
#include "../common/config/config_file.h"
#include "../common/isc_proto.h"
#include "../common/isc_f_proto.h"
#include "../common/utils_proto.h"
#include "../common/ScanDir.h"
#include "../common/os/mod_loader.h"
#include "../common/os/path_utils.h"

#include "Replication.h"
#include "ReplicationUtils.h"

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif

#ifdef WIN_NT
#include <Shellapi.h>
#endif

#include <stdlib.h>
#include <time.h>

//#define USE_CONFIG_FILE

using namespace Firebird;
using namespace Replication;
using namespace ReplicationUtils;

namespace
{
	const char* REPLICATION_CFGFILE = "replication.conf";
	const char* REPLICATION_LOGFILE = "replication.log";

	const char* REPL_CONF_PLUGIN = "replconf";
	const char* REPL_CONF_ENUM_MASTER_FUNC = "enum_repl_master";
	const char* REPL_CONF_ENUM_SLAVE_FUNC = "enum_repl_slave";

	static const ULONG DEFAULT_BUFFER_SIZE = 1024 * 1024; 				// 1 MB
	static const ULONG DEFAULT_LOG_SEGMENT_SIZE = 16 * 1024 * 1024;		// 16 MB
	static const ULONG DEFAULT_LOG_SEGMENT_COUNT = 8;
	static const ULONG DEFAULT_LOG_ARCHIVE_TIMEOUT = 0;
}


// ReplicationUtils namespace

namespace ReplicationUtils
{
	void raiseError(const char* msg, ...)
	{
		char buffer[BUFFER_LARGE];

		va_list ptr;
		va_start(ptr, msg);
		vsprintf(buffer, msg, ptr);
		va_end(ptr);

		Arg::StatusVector error;
		error << Arg::Gds(isc_random) << Arg::Str(buffer);
		error.raise();
	}

	int executeShell(const string& command)
	{
#ifdef WIN_NT
		string params;
		params.printf("/c %s", command.c_str());
		SHELLEXECUTEINFO seInfo = {0};
		seInfo.cbSize = sizeof(SHELLEXECUTEINFO);
		seInfo.fMask = SEE_MASK_NOCLOSEPROCESS;
		seInfo.hwnd = NULL;
		seInfo.lpVerb = NULL;
		seInfo.lpFile = "cmd.exe";
		seInfo.lpParameters = params.c_str();
		seInfo.lpDirectory = NULL;
		seInfo.nShow = SW_HIDE;
		seInfo.hInstApp = NULL;
		ShellExecuteEx(&seInfo);
		WaitForSingleObject(seInfo.hProcess, INFINITE);
		DWORD exitCode = 0;
		GetExitCodeProcess(seInfo.hProcess, &exitCode);
		return (int) exitCode;
#else
		return system(command.c_str());
#endif
	}

	typedef void (*enum_callback_t)(void* arg, const char* key, const char* value);
	typedef int (*enum_func_t)(const char* filter, unsigned errLen, char* errStr,
							   enum_callback_t callback, void* arg);

	typedef GenericMap<Pair<Full<string, string> > > KeyValueMap;

	void callback(void* arg, const char* k, const char* v)
	{
		KeyValueMap* const map = static_cast<KeyValueMap*>(arg);

		const string key(k), value(v);

		string* const stored = map->get(key);

		if (stored)
			stored->assign(value);
		else
			map->put(key, value);
	}

	int enumConfig(const char* database, unsigned errLen, char* errStr,
				   enum_callback_t callback, void* arg, const char* sectionName)
	{
		class PluginError
		{
		public:
			PluginError(unsigned len, char* str)
				: length(len), target(str)
			{}

			int raise(const char* msg, ...)
			{
				char buffer[BUFFER_LARGE];

				va_list ptr;
				va_start(ptr, msg);
				vsprintf(buffer, msg, ptr);
				va_end(ptr);

				strncpy(target, buffer, length);
				return -1;
			}

		private:
			unsigned length;
			char* target;
		};

		PluginError error(errLen, errStr);

		const PathName filename =
			fb_utils::getPrefix(IConfigManager::DIR_CONF, REPLICATION_CFGFILE);

		ConfigFile cfgFile(filename, ConfigFile::HAS_SUB_CONF | ConfigFile::NATIVE_ORDER | ConfigFile::CUSTOM_MACROS);

		int count = 0;
		bool defaultFound = false, exactMatch = false;

		const ConfigFile::Parameters& params = cfgFile.getParameters();
		for (FB_SIZE_T n = 0; n < params.getCount() && !exactMatch; ++n)
		{
			const ConfigFile::Parameter* const section = &params[n];

			if (section->name != sectionName)
				continue;

			const PathName name(section->value.c_str());

			if (!database)
			{
				if (name.hasData())
				{
					(*callback)(arg, name.c_str(), name.c_str());
					count++;
				}

				continue;
			}

			const bool dbEntry = !strcmp(sectionName, "database");
			const bool replEntry = !strcmp(sectionName, "replica");

			if (!dbEntry && !replEntry)
				continue;

			if (name.empty())
			{
				if (defaultFound)
					return error.raise("Only one default section is allowed");

				defaultFound = true;
			}
			else if (database)
			{
				if (name == database)
					exactMatch = true;
				else
					continue;
			}
			else
				exactMatch = true;

			const ConfigFile::Parameters& elements = section->sub->getParameters();
			for (FB_SIZE_T p = 0; p < elements.getCount(); ++p)
			{
				const ConfigFile::Parameter* const el = &elements[p];

				const string key(el->name.c_str());
				const string value(el->value);

				if (value.isEmpty())
					continue;

				if (key == "replica_database" && name.empty())
					error.raise("Target database cannot be specified in the default section");

				if (replEntry &&
					(key == "replica_database" || key == "buffer_size" || key == "disable_on_error" ||
					key == "compress_records" || key == "master_priority" || key == "alert_command" ||
					key == "include_filter" || key == "exclude_filter" || key == "exclude_without_pk" ||
					key == "log_file_prefix" || key == "log_segment_size" || key == "log_segment_count" ||
					key == "log_archive_directory" || key == "log_archive_command" || key == "log_archive_timeout"))
				{
					error.raise("Invalid configuration option: \"%s\"", key.c_str());
				}

				if (dbEntry && (key == "master_database" || key == "db_copy_command" || key == "owner_auth"))
					error.raise("Invalid configuration option: \"%s\"", key.c_str());

				(*callback)(arg, key.c_str(), value.c_str());
				count++;
			}
		}

		return count;
	}

	int enumMasterConfig(const char* database, unsigned errLen, char* errStr,
						 enum_callback_t callback, void* arg)
	{
		return enumConfig(database, errLen, errStr, callback, arg, "database");
	}

	int enumSlaveConfig(const char* database, unsigned errLen, char* errStr,
						enum_callback_t callback, void* arg)
	{
		return enumConfig(database, errLen, errStr, callback, arg, "replica");
	}

	bool getConfig(MemoryPool& pool, Array<ReplicationConfig*>& entries, const PathName& lookup)
	{
		const PathName plugDir = fb_utils::getPrefix(IConfigManager::DIR_PLUGINS, "");
		ScanDir plugins(plugDir.c_str(), "*.*");

		AutoPtr<ModuleLoader::Module> module;
		enum_func_t enum_func = NULL;

		while (plugins.next() && !enum_func)
		{
			const PathName moduleLib(plugins.getFileName());
			if (moduleLib.find(REPL_CONF_PLUGIN) == PathName::npos)
				continue;

			PathName fullModuleName;
			PathUtils::concatPath(fullModuleName, plugDir, moduleLib);

			module = ModuleLoader::loadModule(fullModuleName);
			if (!module)
				raiseError("Replication plugin %s exists but cannot be loaded", fullModuleName.c_str());

			const char* entrypoint =
				lookup.hasData() ? REPL_CONF_ENUM_MASTER_FUNC : REPL_CONF_ENUM_SLAVE_FUNC;
			enum_func = (enum_func_t) module->findSymbol(entrypoint);
		}

		if (!enum_func)
		{
#ifdef USE_CONFIG_FILE
			enum_func = lookup.hasData() ? enumMasterConfig : enumSlaveConfig;
#else
			return false;
#endif
		}

		KeyValueMap map1(pool), map2(pool);
		char errorBuffer[BUFFER_LARGE];

		if ((*enum_func)(NULL, BUFFER_LARGE, errorBuffer, callback, &map1) < 0)
			raiseError(errorBuffer);

		KeyValueMap::ConstAccessor accessor1(&map1);
		if (accessor1.getFirst())
		{
			do {
				const PathName database(accessor1.current()->second.c_str());
				fb_assert(database.hasData());

				if (lookup.hasData() && lookup != database)
					continue;

				if ((*enum_func)(database.c_str(), BUFFER_LARGE, errorBuffer, callback, &map2) < 0)
					raiseError(errorBuffer);

				AutoPtr<ReplicationConfig> config(FB_NEW_POOL(pool) ReplicationConfig(pool));
				config->name = database;

				KeyValueMap::ConstAccessor accessor2(&map2);
				if (accessor2.getFirst())
				{
					do {
						string key = accessor2.current()->first;
						key.trim();
						string value = accessor2.current()->second;
						value.trim();

						char* pEnd = NULL;
						long number = 0;

						if (key.find("replica_database") == 0)
						{
							config->replicas.add(value);
						}
						else if (key == "buffer_size")
						{
							number = strtol(value.c_str(), &pEnd, 10);
							if (pEnd && *pEnd == 0 && number > 0)
								config->bufferSize = (ULONG) number;
						}
						else if (key == "disable_on_error")
						{
							if (value == "true" || value == "yes" || value == "on" || value == "1")
								config->disableOnError = true;
						}
						else if (key == "compress_records")
						{
							if (value == "true" || value == "yes" || value == "on" || value == "1")
								config->compressRecords = true;
						}
						else if (key == "master_priority")
						{
							if (value == "true" || value == "yes" || value == "on" || value == "1")
								config->masterPriority = true;
						}
						else if (key == "alert_command")
						{
							config->alertCommand = value.c_str();
						}
						else if (key == "include_filter")
						{
							config->includeFilter = value;
							ISC_systemToUtf8(config->includeFilter);
						}
						else if (key == "exclude_filter")
						{
							config->excludeFilter = value;
							ISC_systemToUtf8(config->excludeFilter);
						}
						else if (key == "exclude_without_pk")
						{
							if (value == "true" || value == "yes" || value == "on" || value == "1")
								config->excludeWithoutPK = true;
						}
						else if (key == "log_directory")
						{
							config->logDirectory = value.c_str();
							PathUtils::ensureSeparator(config->logDirectory);
						}
						else if (key == "log_file_prefix")
						{
							config->logFilePrefix = value.c_str();
						}
						else if (key == "log_segment_size")
						{
							number = strtol(value.c_str(), &pEnd, 10);
							if (pEnd && *pEnd == 0 && number > 0)
								config->segmentSize = (ULONG) number;
						}
						else if (key == "log_segment_count")
						{
							number = strtol(value.c_str(), &pEnd, 10);
							if (pEnd && *pEnd == 0 && number > 0)
								config->segmentCount = (ULONG) number;
						}
						else if (key == "log_archive_directory")
						{
							config->archiveDirectory = value.c_str();
							PathUtils::ensureSeparator(config->archiveDirectory);
						}
						else if (key == "log_archive_command")
						{
							config->archiveCommand = value.c_str();
						}
						else if (key == "log_archive_timeout")
						{
							number = strtol(value.c_str(), &pEnd, 10);
							if (pEnd && *pEnd == 0 && number > 0)
								config->archiveTimeout = (ULONG) number;
						}
						else if (key == "master_database")
						{
							config->masterDb = value;
						}
						else if (key == "db_copy_command")
						{
							config->copyCommand = value;
						}
						else if (key == "owner_auth")
						{
							config->ownerAuth = value;
						}
						else if (key == "verbose")
						{
							if (value == "true" || value == "yes" || value == "on" || value == "1")
								config->verbose = true;
						}

					} while (accessor2.getNext());
				}

				if (lookup.hasData())
				{
					if (config->replicas.isEmpty() && !config->isLoggingActive())
						raiseError("Neither direct replicas nor logging are configured for replication");

					if (config->isLoggingActive())
					{
						config->compressRecords = true;
						config->masterPriority = true;
					}

					if (config->logFilePrefix.isEmpty())
					{
						PathName db_directory, db_filename;
						PathUtils::splitLastComponent(db_directory, db_filename, database);
						config->logFilePrefix = db_filename;
					}
				}

				entries.add(config.release());

			} while (accessor1.getNext());
		}

		return entries.hasData();
	}
}

// ReplicationConfig class

void ReplicationConfig::get(Array<ReplicationConfig*>& entries)
{
	MemoryPool& pool = *getDefaultMemoryPool();
	getConfig(pool, entries, "");
}

ReplicationConfig* ReplicationConfig::get(const PathName& lookup)
{
	MemoryPool& pool = *getDefaultMemoryPool();
	Array<ReplicationConfig*> entries(pool);
	if (!getConfig(pool, entries, lookup))
		LongJump::raise();
	fb_assert(entries.getCount() == 1);
	return entries[0];
}

ReplicationConfig::ReplicationConfig(MemoryPool& pool)
	: name(pool), replicas(pool),
	  bufferSize(DEFAULT_BUFFER_SIZE),
	  disableOnError(false), compressRecords(false), masterPriority(false),
	  excludeWithoutPK(false),
	  alertCommand(pool), includeFilter(pool), excludeFilter(pool),
	  logDirectory(pool), logFilePrefix(pool),
	  segmentSize(DEFAULT_LOG_SEGMENT_SIZE),
	  segmentCount(DEFAULT_LOG_SEGMENT_COUNT),
	  archiveDirectory(pool), archiveCommand(pool),
	  archiveTimeout(DEFAULT_LOG_ARCHIVE_TIMEOUT),
	  masterDb(pool), copyCommand(pool), ownerAuth(pool),
	  verbose(false)
{}

ReplicationConfig::ReplicationConfig(MemoryPool& pool, const ReplicationConfig& other)
	: name(pool), replicas(pool),
	  bufferSize(other.bufferSize),
	  disableOnError(other.disableOnError),
	  compressRecords(other.compressRecords),
	  masterPriority(other.masterPriority),
	  excludeWithoutPK(other.excludeWithoutPK),
	  alertCommand(pool, other.alertCommand),
	  includeFilter(pool, other.includeFilter),
	  excludeFilter(pool, other.excludeFilter),
	  logDirectory(pool, other.logDirectory),
	  logFilePrefix(pool, other.logFilePrefix),
	  segmentSize(other.segmentSize),
	  segmentCount(other.segmentCount),
	  archiveDirectory(pool, other.archiveDirectory),
	  archiveCommand(pool, other.archiveCommand),
	  archiveTimeout(other.archiveTimeout),
	  masterDb(pool, other.masterDb),
	  copyCommand(pool, other.copyCommand),
	  ownerAuth(pool, other.ownerAuth),
	  verbose(other.verbose)
{}


// ErrorLogger class

#ifdef WIN_NT

ErrorLogger::Guard::Guard()
{
	m_mutex = CreateMutex(NULL, FALSE, "firebird_repl_mutex");
}

ErrorLogger::Guard::~Guard()
{
	CloseHandle(m_mutex);
}

bool ErrorLogger::Guard::lock(FILE* /*file*/)
{
	return (WaitForSingleObject(m_mutex, INFINITE) == WAIT_OBJECT_0);
}

void ErrorLogger::Guard::unlock()
{
	ReleaseMutex(m_mutex);
}

#else

ErrorLogger::Guard::Guard()
{
}

ErrorLogger::Guard::~Guard()
{
}

bool ErrorLogger::Guard::lock(FILE* file)
{
#ifdef HAVE_FLOCK
	if (flock(fileno(file), LOCK_EX))
#else
	if (lockf(fileno(file), F_LOCK, 0))
#endif
	{
		return false;
	}

	return true;
}

void ErrorLogger::Guard::unlock()
{
}

#endif

ErrorLogger::ErrorLogger(MemoryPool& pool, const PathName& database, ReplicationSide side)
	: m_side(side), m_hostname(pool),
	  m_logfile(pool, fb_utils::getPrefix(IConfigManager::DIR_LOG, REPLICATION_LOGFILE)),
	  m_database(pool, database)
{
	char host[BUFFER_LARGE];
	ISC_get_host(host, sizeof(host));
	m_hostname = host;
}

void ErrorLogger::logMessage(const string& message) const
{
	const time_t now = time(NULL);

	static Guard g_guard;

	FILE* const file = fopen(m_logfile.c_str(), "a");
	if (file && g_guard.lock(file))
	{
		const char* const prefix = (m_side == MASTER) ? "master" : "slave";
		fseek(file, 0, SEEK_END);
		fprintf(file, "\n%s (%s) %s\tDatabase: %s\n\t%s\n",
				m_hostname.c_str(), prefix, ctime(&now),
				m_database.c_str(), message.c_str());
		fclose(file);
		g_guard.unlock();
	}
}


// DirectoryWalker class

#if defined(WIN_NT)

DirectoryWalker::DirectoryWalker(const PathName& directory)
	: m_directory(directory)
{
	m_found = TRUE;
	PathName mask = directory + '*';
	m_file = FindFirstFile(mask.c_str(), &m_data);
	if (m_file == INVALID_HANDLE_VALUE)
	{
		if (GetLastError() != ERROR_FILE_NOT_FOUND)
			raiseError("Cannot open directory %s (error %d)", directory.c_str(), ERRNO);

		m_found = FALSE;
	}
	else if (m_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
		++(*this);
}

DirectoryWalker::~DirectoryWalker()
{
	FindClose(m_file);
}

bool DirectoryWalker::hasData() const
{
	return (m_found == TRUE);
}

void DirectoryWalker::operator++()
{
	m_found = FindNextFile(m_file, &m_data);
	if (m_found && (m_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
		++(*this);
}

PathName DirectoryWalker::operator*() const
{
	if (m_found)
	{
		PathName filename(m_directory);
		filename += m_data.cFileName;
		return filename;
	}

	return "";
}

#elif defined(LINUX)

DirectoryWalker::DirectoryWalker(const PathName& directory)
	: m_directory(directory)
{
	m_dir = opendir(directory.c_str());
	if (!m_dir)
		raiseError("Cannot open directory %s (error %d)", directory.c_str(), ERRNO);

	getNext();
}

DirectoryWalker::~DirectoryWalker()
{
	closedir(m_dir);
}

bool DirectoryWalker::hasData() const
{
	return (m_ent != NULL);
}

void DirectoryWalker::operator++()
{
	getNext();
}

PathName DirectoryWalker::operator*() const
{
	if (m_ent)
	{
		PathName filename(m_directory);
		filename += m_ent->d_name;
		return filename;
	}

	return "";
}

void DirectoryWalker::getNext()
{
	while ( (m_ent = readdir(m_dir)) )
	{
		PathName filename(m_directory);
		filename += m_ent->d_name;

		struct stat stats;
		if (!stat(filename.c_str(), &stats) && S_ISREG(stats.st_mode))
			break;
	}
}

#endif
